#include "MeshLoader.h"
#include "CustomHash.h"
#include <iomanip>
#include <algorithm>
#include <list>
#include <set>


std::vector<Node>& MeshLoader::Get_Nodes() {
    return nodes;
}

std::vector<Element>& MeshLoader::Get_Elements() {
    return elements;
}

std::vector<ElemBoundary>& MeshLoader::Get_boundaries() {
    return boundaries;
}


std::vector<int> MeshLoader::Get_Elements_by_ID(int p_node_id1, int p_node_id2, int p_node_id3) {
    std::vector<int> answer;
    auto current_iter = elements.begin();
    auto last_iter = elements.end();
    while (current_iter != last_iter) {
        current_iter = std::find_if(current_iter, last_iter, [p_node_id1, p_node_id2, p_node_id3](const Element& current_element) {
            return std::find(current_element.node_id.begin(), current_element.node_id.end(), p_node_id1) != current_element.node_id.end()
                && std::find(current_element.node_id.begin(), current_element.node_id.end(), p_node_id2) != current_element.node_id.end()
                && std::find(current_element.node_id.begin(), current_element.node_id.end(), p_node_id3) != current_element.node_id.end(); });
        
        if (current_iter != last_iter) {
            answer.push_back(current_iter->element_id);
            ++current_iter;
        }
    }
    return answer;
}

std::vector<int> MeshLoader::Get_Elements_by_edge(int id1, int id2) {
    std::vector<int> answer;

    auto current_iter = elements.begin();
    auto last_iter = elements.end();
    while (current_iter != last_iter) {
        current_iter = std::find_if(current_iter, last_iter, [id1, id2](Element current_element) {
            return std::find(current_element.node_id.begin(), current_element.node_id.end(), id1) != current_element.node_id.end()
                && std::find(current_element.node_id.begin(), current_element.node_id.end(), id2) != current_element.node_id.end(); });
        if (current_iter != last_iter) {
            answer.push_back(current_iter->element_id);
            ++current_iter;
        }
    }
    return answer;
}


std::vector<int> MeshLoader::Get_elems_by_material(int p_material) const {
    std::vector<int> result;
    auto first_element = elements.begin();
    auto last_element = elements.end();
    while (first_element != last_element) {
        auto found_elem = std::find_if(first_element, last_element,
            [p_material](const Element& elem) 
            {
                return elem.material_id == p_material;
            });

        first_element = found_elem;
        if (first_element != last_element) {
            result.push_back(first_element->element_id);
            ++first_element;
        }
    }
    return result;
}


std::set<int> MeshLoader::Get_nodes_by_surface(int p_surface) {
    std::set<int> result;
    auto first = boundaries.begin();
    auto last = boundaries.end();
    while (first != last) {
        auto found = std::find_if(first, last,
            [p_surface](const ElemBoundary& cur_boundary) {
                return cur_boundary.surface_id == p_surface;
            });
        first = found;
        if (found != last) {
            for (const auto& node_in_boundary : found->node_id)
                result.insert(node_in_boundary);
            ++first;
        }

    }
    return result;
}

std::vector<std::set<int>> MeshLoader::Get_Nodes_Neighbors() const {
    std::vector<std::set<int>> result(nodes.size() + 1); 
    for (const auto& cur_elem : elements) {
        for (auto cur_node_id : cur_elem.node_id)
            for (auto adj_node_id : cur_elem.node_id)
                if (cur_node_id != adj_node_id)
                    result[cur_node_id].insert(adj_node_id);
    }

    return result;

}

Node MeshLoader::Get_Node_by_ID(int p_id) {
    return nodes.at(p_id);
}

std::vector<int> MeshLoader::Get_boundaries_by_surface(int p_surface) const
{
    std::vector<int> res;

    auto first = boundaries.begin();
    auto last = boundaries.end();

    while (first != last) {
        auto found_elem = std::find_if(first, last,
            [p_surface](const ElemBoundary& elem) {
                return elem.surface_id == p_surface;
            });

        first = found_elem;
        if (found_elem != last) {
            res.push_back(found_elem->element_id);
            ++first;
        }
    }

    return res;
}

std::array<double, 3> MeshLoader::Get_ords_node_by_ID(int p_id)
{
    return nodes.at(p_id).ords;
}

Node& MeshLoader::Get_node_by_ID(int p_id)
{
    return nodes.at(p_id);
}

std::array<double, 3> MeshLoader::Get_middle_of_edge(const Edge& p_edge)
{
    return
    { (get<0>(nodes[p_edge.nodes.first - 1].ords) + get<0>(nodes[p_edge.nodes.second - 1].ords)) * 1 / 2
    , (get<1>(nodes[p_edge.nodes.first - 1].ords) + get<1>(nodes[p_edge.nodes.second - 1].ords)) * 1 / 2
    , (get<2>(nodes[p_edge.nodes.first - 1].ords) + get<2>(nodes[p_edge.nodes.second - 1].ords)) * 1 / 2 };
}


Node MeshLoader::Get_middle_node(const Edge& cur_edge)
{
    Node new_node(nodes.size() + 1, Get_middle_of_edge(cur_edge), 0);

    return new_node;
}

void MeshLoader::Insert_newNodes()
{
    
    std::unordered_set <Edge> examined_edges;

    for (auto& elem : elements) {
        std::vector<int> cur_nodes_id = elem.node_id;
     
        for (auto first_node = 0; first_node < 4; ++first_node)
            for (auto second_node = first_node + 1; second_node < 4; ++second_node) {
                Edge cur_edge(cur_nodes_id[first_node], cur_nodes_id[second_node]);

                if (examined_edges.insert(cur_edge).second) { 
                    Node new_node = Get_middle_node(cur_edge);
                    cur_edge.update_mid(new_node.id);
                    nodes.push_back(new_node);
                }
                else {
                    elem.node_id.push_back(cur_edge.middle_node);
                }
            }
    }


    for (auto& elem : boundaries) {
        std::vector<int> cur_nodes_id = elem.node_id;
        for (auto first_node = 0; first_node < 3; ++first_node)
            for (auto second_node = first_node + 1; second_node < 3; ++second_node) {
                auto found_edge = examined_edges.find({ cur_nodes_id[first_node], cur_nodes_id[second_node] });
                if (found_edge == examined_edges.end())
                    found_edge = examined_edges.find({ cur_nodes_id[second_node], cur_nodes_id[first_node] });
                elem.node_id.push_back(found_edge->middle_node);
            }
    }
}

